/*
 * Copyright (C) 2012-2014  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <libc64_dsp/include/inc_overlay.h>

#include "scalers.h"


/* 2xSaI functions */
#define BITFORMAT 565
//#define BITFORMAT 555

#if (BITFORMAT == 565)

#define colorMask 0xF7DEF7DE
#define lowPixelMask 0x08210821
#define qcolorMask 0xE79CE79C
#define qlowpixelMask 0x18631863

#elif (BITFORMAT == 555)

#define colorMask 0x7BDE7BDE
#define lowPixelMask 0x04210421
#define qcolorMask 0x739C739C
#define qlowpixelMask 0x0C630C63

#endif


static inline uint32_t INTERPOLATE(uint32_t A, uint32_t B)
{
    return ( ((A & colorMask) >> 1) + ((B & colorMask) >> 1) + (A & B & lowPixelMask) );
}


static inline uint32_t Q_INTERPOLATE(uint32_t A, uint32_t B, uint32_t C, uint32_t D)
{
    register uint32_t x = ((A & qcolorMask) >> 2) +
                          ((B & qcolorMask) >> 2) +
                          ((C & qcolorMask) >> 2) +
                          ((D & qcolorMask) >> 2);
    register uint32_t y = (A & qlowpixelMask) +
                          (B & qlowpixelMask) +
                          (C & qlowpixelMask) +
                          (D & qlowpixelMask);
    y = (y>>2) & qlowpixelMask;
    return x+y;
}


static void __2xsai_16_16_line(const uint16_t * restrict src1, const uint16_t * restrict src2, const uint16_t * restrict src3, const uint16_t * restrict src4, int width, uint16_t * restrict dst1, uint16_t * restrict dst2)
{
    register uint32_t colorA, colorB;
    uint32_t colorC, colorD,
             colorE, colorF, colorG, colorH,
             colorI, colorJ, colorK, colorL,
             colorM, colorN, colorO/*, colorP*/;
    uint32_t product, product1, product2;
    uint32_t Mask1, Mask2, Mask3, r;

    _nassert((int)src1 % 8 == 0);
    _nassert((int)src2 % 8 == 0);
    _nassert((int)src3 % 8 == 0);
    _nassert((int)src4 % 8 == 0);
    _nassert((int)dst1 % 8 == 0);
    _nassert((int)dst2 % 8 == 0);
    _nassert(width >= 32);
    _nassert(width % 4 == 0);


    colorJ = _amem4_const(src1);
    src1+=2;
    colorF = _pack2(colorJ, colorJ);

    colorK = _amem4_const(src2);
    src2+=2;
    colorB = _pack2(colorK, colorK);

    colorL = _amem4_const(src3);
    src3+=2;
    colorD = _pack2(colorL, colorL);

    colorO = *src4;
    src4++;
    colorO = _pack2(colorO, colorO);


    for (width -= 2; width != 0; width -= 2)
    {
//---------------------------------------
// Map of the pixels:                    I|E F|J
//                                       G|A B|K
//                                       H|C D|L
//                                       M|N O|P
        colorI = colorF;
        colorE = colorJ;
        colorJ = _amem4_const(src1);
        src1+=2;
        colorF = _packlh2(colorJ, colorE);

        colorG = colorB;
        colorA = colorK;
        colorK = _amem4_const(src2);
        src2+=2;
        colorB = _packlh2(colorK, colorA);

        colorH = colorD;
        colorC = colorL;
        colorL = _amem4_const(src3);
        src3+=2;
        colorD = _packlh2(colorL, colorC);

        colorM = colorO;
        colorO = _mem4_const(src4);
        src4+=2;
        colorN = _packlh2(colorO, colorM);


        Mask1 = _xpnd2( ( _cmpeq2(colorA, colorD) & (~_cmpeq2(colorB, colorC)) &   _cmpeq2(colorA, colorE)  & _cmpeq2(colorB, colorL) ) |
                        ( _cmpeq2(colorA, colorC) &   _cmpeq2(colorA, colorF)  & (~_cmpeq2(colorB, colorE)) & _cmpeq2(colorB, colorJ) ) );

        Mask2 = _xpnd2( ( _cmpeq2(colorB, colorC) & (~_cmpeq2(colorA, colorD)) &   _cmpeq2(colorB, colorF)  & _cmpeq2(colorA, colorH) ) |
                        ( _cmpeq2(colorB, colorE) &   _cmpeq2(colorB, colorD)  & (~_cmpeq2(colorA, colorF)) & _cmpeq2(colorA, colorI) ) );

        product = (colorA & Mask1) | (colorB & Mask2) | (INTERPOLATE(colorA, colorB) & ~(Mask1 | Mask2));

        _amem8(dst1) = _dpack2(product, colorA);
        dst1+=4;


        Mask1 = _xpnd2( ( _cmpeq2(colorA, colorD) & (~_cmpeq2(colorB, colorC)) &   _cmpeq2(colorA, colorG)  & _cmpeq2(colorC, colorO) ) |
                        ( _cmpeq2(colorA, colorB) &   _cmpeq2(colorA, colorH)  & (~_cmpeq2(colorG, colorC)) & _cmpeq2(colorC, colorM) ) );

        Mask2 = _xpnd2( ( _cmpeq2(colorB, colorC) & (~_cmpeq2(colorA, colorD)) &   _cmpeq2(colorC, colorH)  & _cmpeq2(colorA, colorF) ) |
                        ( _cmpeq2(colorC, colorG) &   _cmpeq2(colorC, colorD)  & (~_cmpeq2(colorA, colorH)) & _cmpeq2(colorA, colorI) ) );

        product1 = (colorA & Mask1) | (colorC & Mask2) | (INTERPOLATE(colorA, colorC) & ~(Mask1 | Mask2));


        Mask1 = _xpnd2( ( _cmpeq2(colorA, colorD) & (~_cmpeq2(colorB, colorC)) ) |
                        ( _cmpeq2(colorA, colorD) &   _cmpeq2(colorB, colorC)  & _cmpeq2(colorA, colorB) ) );

        Mask2 = _xpnd2( ( _cmpeq2(colorB, colorC) & (~_cmpeq2(colorA, colorD)) ) |
                        ( _cmpeq2(colorB, colorC) &   _cmpeq2(colorA, colorD)  & _cmpeq2(colorA, colorB) ) );


        Mask3 = _xpnd2( _cmpeq2(colorA, colorD) & _cmpeq2(colorB, colorC) & (~_cmpeq2(colorA, colorB)) );

        if ( Mask3 )
        {
            r =          _xpnd2( _cmpeq2(colorA, colorG) & _cmpeq2(colorA, colorE) );
            r = _sub2(r, _xpnd2( _cmpeq2(colorB, colorG) & _cmpeq2(colorB, colorE) & (~_cmpeq2(colorA, colorG)) & (~_cmpeq2(colorA, colorE)) ) );

            r = _sub2(r, _xpnd2( _cmpeq2(colorB, colorK) & _cmpeq2(colorB, colorF) ) );
            r = _add2(r, _xpnd2( _cmpeq2(colorA, colorK) & _cmpeq2(colorA, colorF) & (~_cmpeq2(colorB, colorK)) & (~_cmpeq2(colorB, colorF)) ) );

            r = _sub2(r, _xpnd2( _cmpeq2(colorB, colorH) & _cmpeq2(colorB, colorN) ) );
            r = _add2(r, _xpnd2( _cmpeq2(colorA, colorH) & _cmpeq2(colorA, colorN) & (~_cmpeq2(colorB, colorH)) & (~_cmpeq2(colorB, colorN)) ) );

            r = _add2(r, _xpnd2( _cmpeq2(colorA, colorL) & _cmpeq2(colorA, colorO) ) );
            r = _sub2(r, _xpnd2( _cmpeq2(colorB, colorL) & _cmpeq2(colorB, colorO) & (~_cmpeq2(colorA, colorL)) & (~_cmpeq2(colorA, colorO)) ) );

            Mask1 |= _xpnd2( _cmpgt2(r, 0) ) & Mask3;
            Mask2 |= _xpnd2( _cmpgt2(0, r) ) & Mask3;
        }

        product2 = (colorA & Mask1) | (colorB & Mask2) | (Q_INTERPOLATE(colorA, colorB, colorC, colorD) & ~(Mask1 | Mask2));

        _amem8(dst2) = _dpack2(product2, product1);
        dst2+=4;

    }

    // last pixels
    {
        colorI = colorF;
        colorE = colorJ;
        colorF = _packh2(colorJ, colorJ);
        colorJ = colorF;

        colorG = colorB;
        colorA = colorK;
        colorB = _packh2(colorK, colorK);
        colorK = colorB;

        colorH = colorD;
        colorC = colorL;
        colorD = _packh2(colorL, colorL);
        colorL = colorD;

        colorM = colorO;
        colorN = _packlh2(*src4, colorO);
        colorO = _packh2(colorN, colorN);


        Mask1 = _xpnd2( ( _cmpeq2(colorA, colorD) & (~_cmpeq2(colorB, colorC)) &   _cmpeq2(colorA, colorE)  & _cmpeq2(colorB, colorL) ) |
                        ( _cmpeq2(colorA, colorC) &   _cmpeq2(colorA, colorF)  & (~_cmpeq2(colorB, colorE)) & _cmpeq2(colorB, colorJ) ) );

        Mask2 = _xpnd2( ( _cmpeq2(colorB, colorC) & (~_cmpeq2(colorA, colorD)) &   _cmpeq2(colorB, colorF)  & _cmpeq2(colorA, colorH) ) |
                        ( _cmpeq2(colorB, colorE) &   _cmpeq2(colorB, colorD)  & (~_cmpeq2(colorA, colorF)) & _cmpeq2(colorA, colorI) ) );

        product = (colorA & Mask1) | (colorB & Mask2) | (INTERPOLATE(colorA, colorB) & ~(Mask1 | Mask2));

        _amem8(dst1) = _dpack2(product, colorA);
        dst1+=4;


        Mask1 = _xpnd2( ( _cmpeq2(colorA, colorD) & (~_cmpeq2(colorB, colorC)) &   _cmpeq2(colorA, colorG)  & _cmpeq2(colorC, colorO) ) |
                        ( _cmpeq2(colorA, colorB) &   _cmpeq2(colorA, colorH)  & (~_cmpeq2(colorG, colorC)) & _cmpeq2(colorC, colorM) ) );

        Mask2 = _xpnd2( ( _cmpeq2(colorB, colorC) & (~_cmpeq2(colorA, colorD)) &   _cmpeq2(colorC, colorH)  & _cmpeq2(colorA, colorF) ) |
                        ( _cmpeq2(colorC, colorG) &   _cmpeq2(colorC, colorD)  & (~_cmpeq2(colorA, colorH)) & _cmpeq2(colorA, colorI) ) );

        product1 = (colorA & Mask1) | (colorC & Mask2) | (INTERPOLATE(colorA, colorC) & ~(Mask1 | Mask2));


        Mask1 = _xpnd2( ( _cmpeq2(colorA, colorD) & (~_cmpeq2(colorB, colorC)) ) |
                        ( _cmpeq2(colorA, colorD) &   _cmpeq2(colorB, colorC)  & _cmpeq2(colorA, colorB) ) );

        Mask2 = _xpnd2( ( _cmpeq2(colorB, colorC) & (~_cmpeq2(colorA, colorD)) ) |
                        ( _cmpeq2(colorB, colorC) &   _cmpeq2(colorA, colorD)  & _cmpeq2(colorA, colorB) ) );


        Mask3 = _xpnd2( _cmpeq2(colorA, colorD) & _cmpeq2(colorB, colorC) & (~_cmpeq2(colorA, colorB)) );

        if ( Mask3 )
        {
            r =          _xpnd2( _cmpeq2(colorA, colorG) & _cmpeq2(colorA, colorE) );
            r = _sub2(r, _xpnd2( _cmpeq2(colorB, colorG) & _cmpeq2(colorB, colorE) & (~_cmpeq2(colorA, colorG)) & (~_cmpeq2(colorA, colorE)) ) );

            //r = _sub2(r, _xpnd2( _cmpeq2(colorB, colorK) & _cmpeq2(colorB, colorF) ) );
            //r = _add2(r, _xpnd2( _cmpeq2(colorA, colorK) & _cmpeq2(colorA, colorF) & (~_cmpeq2(colorB, colorK)) & (~_cmpeq2(colorB, colorF)) ) );
            r = _sub2(r, _xpnd2( _cmpeq2(colorB, colorF) ) );

            r = _sub2(r, _xpnd2( _cmpeq2(colorB, colorH) & _cmpeq2(colorB, colorN) ) );
            r = _add2(r, _xpnd2( _cmpeq2(colorA, colorH) & _cmpeq2(colorA, colorN) & (~_cmpeq2(colorB, colorH)) & (~_cmpeq2(colorB, colorN)) ) );

            r = _add2(r, _xpnd2( _cmpeq2(colorA, colorL) & _cmpeq2(colorA, colorO) ) );
            r = _sub2(r, _xpnd2( _cmpeq2(colorB, colorL) & _cmpeq2(colorB, colorO) & (~_cmpeq2(colorA, colorL)) & (~_cmpeq2(colorA, colorO)) ) );

            Mask1 |= _xpnd2( _cmpgt2(r, 0) ) & Mask3;
            Mask2 |= _xpnd2( _cmpgt2(0, r) ) & Mask3;
        }

        product2 = (colorA & Mask1) | (colorB & Mask2) | (Q_INTERPOLATE(colorA, colorB, colorC, colorD) & ~(Mask1 | Mask2));

        _amem8(dst2) = _dpack2(product2, product1);
        dst2+=4;

    }
}

void __2xsai_16_16(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 5);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    __2xsai_16_16_line(src, src, src + (srcstride / 2), src + srcstride, width, dst, dst + (dststride / 2));

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    for (height -= 3; height != 0; height--)
    {
        __2xsai_16_16_line(src - (srcstride / 2), src, src + (srcstride / 2), src + srcstride, width, dst, dst + (dststride / 2));

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    __2xsai_16_16_line(src - (srcstride / 2), src, src + (srcstride / 2), src + (srcstride / 2), width, dst, dst + (dststride / 2));

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    __2xsai_16_16_line(src - (srcstride / 2), src, src, src, width, dst, dst + (dststride / 2));
}

void __2xsai_16_16_srcdma(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 5);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict tmpsrc1;
    uint16_t * restrict tmpsrc2;
    uint16_t * restrict tmpsrc3;
    uint16_t * restrict tmpsrc4;
    uint16_t * restrict tmpsrc5;

    tmpsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    syscalls.qdma_copy2d(QDMA_0, tmpsrc1, 2*width, src, srcstride, 2*width, 3);

    tmpsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    tmpsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);
    tmpsrc4 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);
    tmpsrc5 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src + (3 * srcstride / 2), 2*width);

    __2xsai_16_16_line(tmpsrc1, tmpsrc1, tmpsrc2, tmpsrc3, width, dst, dst + (dststride / 2));

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    for (height -= 4; height != 0; height--)
    {
        syscalls.qdma_wait(QDMA_0);
        syscalls.qdma_copy1d(QDMA_0, tmpsrc5, src + (3 * srcstride / 2), 2*width);
        __2xsai_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpsrc4, width, dst, dst + (dststride / 2));

        {
            uint16_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc5;
            tmpsrc5 = tmpsrc;
        }

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    syscalls.qdma_wait(QDMA_0);
    __2xsai_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpsrc4, width, dst, dst + (dststride / 2));

    //src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    __2xsai_16_16_line(tmpsrc2, tmpsrc3, tmpsrc4, tmpsrc4, width, dst, dst + (dststride / 2));

    //src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    __2xsai_16_16_line(tmpsrc3, tmpsrc4, tmpsrc4, tmpsrc4, width, dst, dst + (dststride / 2));
}

void __2xsai_16_16_dstdma(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 5);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    tmpdst1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    tmpdst2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);

    __2xsai_16_16_line(src, src, src + (srcstride / 2), src + srcstride, width, tmpdst2, tmpdst2 + (2*width));

    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    for (height -= 3; height != 0; height--)
    {
        __2xsai_16_16_line(src - (srcstride / 2), src, src + (srcstride / 2), src + srcstride, width, tmpdst1, tmpdst1 + (2*width));

        syscalls.qdma_wait(QDMA_1);
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);

        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    __2xsai_16_16_line(src - (srcstride / 2), src, src + (srcstride / 2), src + (srcstride / 2), width, tmpdst1, tmpdst1 + (2*width));

    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    __2xsai_16_16_line(src - (srcstride / 2), src, src, src, width, tmpdst2, tmpdst2 + (2*width));

    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst2, 4*width, 4*width, 2);
    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_wait(QDMA_0);
}

void __2xsai_16_16_dma(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 5);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict tmpsrc1;
    uint16_t * restrict tmpsrc2;
    uint16_t * restrict tmpsrc3;
    uint16_t * restrict tmpsrc4;
    uint16_t * restrict tmpsrc5;
    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    tmpsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    syscalls.qdma_copy2d(QDMA_0, tmpsrc1, 2*width, src, srcstride, 2*width, 3);

    tmpsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    tmpsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);
    tmpsrc4 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);
    tmpsrc5 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);
    tmpdst1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 10*width);
    tmpdst2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 18*width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src + (3 * srcstride / 2), 2*width);

    __2xsai_16_16_line(tmpsrc1, tmpsrc1, tmpsrc2, tmpsrc3, width, tmpdst2, tmpdst2 + (2*width));

    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    for (height -= 4; height != 0; height--)
    {
        syscalls.qdma_wait(QDMA_0);
        syscalls.qdma_copy1d(QDMA_0, tmpsrc5, src + (3 * srcstride / 2), 2*width);
        __2xsai_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpsrc4, width, tmpdst1, tmpdst1 + (2*width));

        syscalls.qdma_wait(QDMA_1);
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);

        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc5;
            tmpsrc5 = tmpsrc;
        }

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    syscalls.qdma_wait(QDMA_0);
    __2xsai_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpsrc4, width, tmpdst1, tmpdst1 + (2*width));

    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);

    //src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    __2xsai_16_16_line(tmpsrc2, tmpsrc3, tmpsrc4, tmpsrc4, width, tmpdst2, tmpdst2 + (2*width));

    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    //src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    __2xsai_16_16_line(tmpsrc3, tmpsrc4, tmpsrc4, tmpsrc4, width, tmpdst1, tmpdst1 + (2*width));

    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_wait(QDMA_0);
}

void __2xsai_16_16_dmalink(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 5);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict tmpsrc1;
    uint16_t * restrict tmpsrc2;
    uint16_t * restrict tmpsrc3;
    uint16_t * restrict tmpsrc4;
    uint16_t * restrict tmpsrc5;
    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    tmpsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    syscalls.qdma_copy2d(QDMA_0, tmpsrc1, 2*width, src, srcstride, 2*width, 3);

    tmpsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    tmpsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);
    tmpsrc4 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);
    tmpsrc5 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);
    tmpdst1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 10*width);
    tmpdst2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 18*width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src + (3 * srcstride / 2), 2*width);

    __2xsai_16_16_line(tmpsrc1, tmpsrc1, tmpsrc2, tmpsrc3, width, tmpdst2, tmpdst2 + (2*width));


    src += (srcstride / 2);

    for (height -= 4; height != 0; height--)
    {
        syscalls.qdma_wait(QDMA_0);
        syscalls.qdma_link2d(QDMA_0, 8, 0xFFFF, dst, dststride, tmpdst2, 4*width, 4*width, 2);
        syscalls.qdma_link1d(QDMA_0, QDMA_0, 8, tmpsrc5, src + (3 * srcstride / 2), 2*width);

        __2xsai_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpsrc4, width, tmpdst1, tmpdst1 + (2*width));

        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc5;
            tmpsrc5 = tmpsrc;
        }

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    __2xsai_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpsrc4, width, tmpdst1, tmpdst1 + (2*width));

    dst += 2 * (dststride / 2);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);

    __2xsai_16_16_line(tmpsrc2, tmpsrc3, tmpsrc4, tmpsrc4, width, tmpdst2, tmpdst2 + (2*width));

    dst += 2 * (dststride / 2);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    __2xsai_16_16_line(tmpsrc3, tmpsrc4, tmpsrc4, tmpsrc4, width, tmpdst1, tmpdst1 + (2*width));

    dst += 2 * (dststride / 2);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);

    syscalls.qdma_wait(QDMA_0);
}

