/*
 * Copyright (C) 2012-2014  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <libc64_dsp/include/inc_overlay.h>

#include "scalers.h"


static void _eagle2x_16_16_line(const uint16_t * restrict src1, const uint16_t * restrict src2, const uint16_t * restrict src3, uint16_t * restrict dst1, uint16_t * restrict dst2, int width)
{
    uint64_t cU2, cW2, cZ2;
    uint32_t cT, cC, cY, cS, cV, cX;
    uint32_t tmp1, tmp2, tmp3, tmp4, E1, E2, E3, E4;

    _nassert((int)src1 % 8 == 0);
    _nassert((int)src2 % 8 == 0);
    _nassert((int)src3 % 8 == 0);
    _nassert((int)dst1 % 8 == 0);
    _nassert((int)dst2 % 8 == 0);
    _nassert(width >= 32);
    _nassert(width % 4 == 0);

    cU2 = *src1;
    src1++;
    cU2 = ((uint64_t) _pack2(cU2, cU2)) << 32;

    cW2 = *src2;
    src2++;
    cW2 = ((uint64_t) _pack2(cW2, cW2)) << 32;

    cZ2 = *src3;
    src3++;
    cZ2 = ((uint64_t) _pack2(cZ2, cZ2)) << 32;

    // inner loop (4 pixels per iteration)
    #pragma MUST_ITERATE(7, , )
    for (width -= 4; width != 0; width -= 4)
    {
// S T U  --\ E1 E2
// V C W  --/ E3 E4
// X Y Z

        cS = _hill(cU2);
        cU2 = _mem8_const(src1);
        src1 += 4;

        cV = _hill(cW2);
        cW2 = _mem8_const(src2);
        src2 += 4;

        cX = _hill(cZ2);
        cZ2 = _mem8_const(src3);
        src3 += 4;


        #define cU _loll(cU2)
        #define cW _loll(cW2)
        #define cZ _loll(cZ2)

        cT = _packlh2(cU, cS);
        cC = _packlh2(cW, cV);
        cY = _packlh2(cZ, cX);

        tmp1 = _xpnd2(_cmpeq2(cS, cT) & _cmpeq2(cS, cV));   // tmp1 = S == T && S == V
        tmp2 = _xpnd2(_cmpeq2(cU, cT) & _cmpeq2(cU, cW));   // tmp2 = U == T && U == W

        E1 = (cS & tmp1) | (cC & ~tmp1);                    // E1 = tmp1 ? S : C
        E2 = (cU & tmp2) | (cC & ~tmp2);                    // E2 = tmp2 ? U : C

        _amem8(dst1) = _dpack2(E2, E1);
        dst1 += 4;


        tmp3 = _xpnd2(_cmpeq2(cX, cV) & _cmpeq2(cX, cY));   // tmp3 = X == V && X == Y
        tmp4 = _xpnd2(_cmpeq2(cZ, cW) & _cmpeq2(cZ, cY));   // tmp4 = Z == W && Z == Y

        E3 = (cX & tmp3) | (cC & ~tmp3);                    // E3 = tmp3 ? X : C
        E4 = (cZ & tmp4) | (cC & ~tmp4);                    // E4 = tmp4 ? Z : C

        _amem8(dst2) = _dpack2(E4, E3);
        dst2 += 4;

        #undef cZ
        #undef cW
        #undef cU


        #define cU _hill(cU2)
        #define cW _hill(cW2)
        #define cZ _hill(cZ2)
        #define cS _loll(cU2)
        #define cV _loll(cW2)
        #define cX _loll(cZ2)


        cT = _packlh2(cU, cS);
        cC = _packlh2(cW, cV);
        cY = _packlh2(cZ, cX);

        tmp1 = _xpnd2(_cmpeq2(cS, cT) & _cmpeq2(cS, cV));   // tmp1 = S == T && S == V
        tmp2 = _xpnd2(_cmpeq2(cU, cT) & _cmpeq2(cU, cW));   // tmp2 = U == T && U == W

        E1 = (cS & tmp1) | (cC & ~tmp1);                    // E1 = tmp1 ? S : C
        E2 = (cU & tmp2) | (cC & ~tmp2);                    // E2 = tmp2 ? U : C

        _amem8(dst1) = _dpack2(E2, E1);
        dst1 += 4;


        tmp3 = _xpnd2(_cmpeq2(cX, cV) & _cmpeq2(cX, cY));   // tmp3 = X == V && X == Y
        tmp4 = _xpnd2(_cmpeq2(cZ, cW) & _cmpeq2(cZ, cY));   // tmp4 = Z == W && Z == Y

        E3 = (cX & tmp3) | (cC & ~tmp3);                    // E3 = tmp3 ? X : C
        E4 = (cZ & tmp4) | (cC & ~tmp4);                    // E4 = tmp4 ? Z : C

        _amem8(dst2) = _dpack2(E4, E3);
        dst2 += 4;

        #undef cX
        #undef cV
        #undef cS
        #undef cZ
        #undef cW
        #undef cU
    }

    // last 4 pixels
    {
        cS = _hill(cU2);
        cU2 = _itoll(_pack2(src1[2], src1[2]), _mem4_const(src1));
        //src1 += 4;

        cV = _hill(cW2);
        cW2 = _itoll(_pack2(src2[2], src2[2]), _mem4_const(src2));
        //src2 += 4;

        cX = _hill(cZ2);
        cZ2 = _itoll(_pack2(src3[2], src3[2]), _mem4_const(src3));
        //src3 += 4;


        #define cU _loll(cU2)
        #define cW _loll(cW2)
        #define cZ _loll(cZ2)

        cT = _packlh2(cU, cS);
        cC = _packlh2(cW, cV);
        cY = _packlh2(cZ, cX);

        tmp1 = _xpnd2(_cmpeq2(cS, cT) & _cmpeq2(cS, cV));   // tmp1 = S == T && S == V
        tmp2 = _xpnd2(_cmpeq2(cU, cT) & _cmpeq2(cU, cW));   // tmp2 = U == T && U == W

        E1 = (cS & tmp1) | (cC & ~tmp1);                    // E1 = tmp1 ? S : C
        E2 = (cU & tmp2) | (cC & ~tmp2);                    // E2 = tmp2 ? U : C

        _amem8(dst1) = _dpack2(E2, E1);
        dst1 += 4;


        tmp3 = _xpnd2(_cmpeq2(cX, cV) & _cmpeq2(cX, cY));   // tmp3 = X == V && X == Y
        tmp4 = _xpnd2(_cmpeq2(cZ, cW) & _cmpeq2(cZ, cY));   // tmp4 = Z == W && Z == Y

        E3 = (cX & tmp3) | (cC & ~tmp3);                    // E3 = tmp3 ? X : C
        E4 = (cZ & tmp4) | (cC & ~tmp4);                    // E4 = tmp4 ? Z : C

        _amem8(dst2) = _dpack2(E4, E3);
        dst2 += 4;

        #undef cZ
        #undef cW
        #undef cU


        #define cU _hill(cU2)
        #define cW _hill(cW2)
        #define cZ _hill(cZ2)
        #define cS _loll(cU2)
        #define cV _loll(cW2)
        #define cX _loll(cZ2)


        cT = _packlh2(cU, cS);
        cC = _packlh2(cW, cV);
        cY = _packlh2(cZ, cX);

        tmp1 = _xpnd2(_cmpeq2(cS, cT) & _cmpeq2(cS, cV));   // tmp1 = S == T && S == V
        tmp2 = _xpnd2(_cmpeq2(cU, cT) & _cmpeq2(cU, cW));   // tmp2 = U == T && U == W

        E1 = (cS & tmp1) | (cC & ~tmp1);                    // E1 = tmp1 ? S : C
        E2 = (cU & tmp2) | (cC & ~tmp2);                    // E2 = tmp2 ? U : C

        _amem8(dst1) = _dpack2(E2, E1);
        //dst1 += 4;


        tmp3 = _xpnd2(_cmpeq2(cX, cV) & _cmpeq2(cX, cY));   // tmp3 = X == V && X == Y
        tmp4 = _xpnd2(_cmpeq2(cZ, cW) & _cmpeq2(cZ, cY));   // tmp4 = Z == W && Z == Y

        E3 = (cX & tmp3) | (cC & ~tmp3);                    // E3 = tmp3 ? X : C
        E4 = (cZ & tmp4) | (cC & ~tmp4);                    // E4 = tmp4 ? Z : C

        _amem8(dst2) = _dpack2(E4, E3);
        //dst2 += 4;

        #undef cX
        #undef cV
        #undef cS
        #undef cZ
        #undef cW
        #undef cU
    }
}

void _eagle2x_16_16(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    _eagle2x_16_16_line(src, src, src + (srcstride / 2), dst, dst + (dststride / 2), width);

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    for (height -= 2; height != 0; height--)
    {
        _eagle2x_16_16_line(src - (srcstride / 2), src, src + (srcstride / 2), dst, dst + (dststride / 2), width);

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    _eagle2x_16_16_line(src - (srcstride / 2), src, src, dst, dst + (dststride / 2), width);
}

