/*
 * Copyright (C) 2012-2014  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <libc64_dsp/include/inc_overlay.h>

#include "scalers.h"


static void _normal1x_8_16_line(const uint8_t * restrict src, uint16_t * restrict dst, const uint16_t * restrict pal, int width)
{
    uint64_t tmp;

    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert((int)pal % 8 == 0);
    _nassert(width >= 32);
    _nassert(width % 8 == 0);

    #pragma MUST_ITERATE(4, ,)
    for (; width != 0; width -= 8)
    {
        tmp = _amem8_const(src);
        src += 8;

        _amem8(dst) = pal[(_loll(tmp)      ) & 0xff] | (((uint32_t)pal[(_loll(tmp) >>  8) & 0xff]) << 16) | (((uint64_t)(pal[(_loll(tmp) >> 16) & 0xff] | (((uint32_t)pal[(_loll(tmp) >> 24)       ]) << 16))) << 32);
        dst += 4;
        _amem8(dst) = pal[(_hill(tmp)      ) & 0xff] | (((uint32_t)pal[(_hill(tmp) >>  8) & 0xff]) << 16) | (((uint64_t)(pal[(_hill(tmp) >> 16) & 0xff] | (((uint32_t)pal[(_hill(tmp) >> 24)       ]) << 16))) << 32);
        dst += 4;
    }
}


/* Scale2x functions */
static void _scale2x_16_16_line(const uint16_t * restrict src1, const uint16_t * restrict src2, const uint16_t * restrict src3, uint16_t * restrict dst1, uint16_t * restrict dst2, int width)
{
    uint64_t cB2, cD2, cF2, cH2;
    uint32_t cE, cnewD;
    uint32_t C0, tmp0, tmp1, tmp2, tmp3, E0, E1, E2, E3;

    _nassert((int)src1 % 8 == 0);
    _nassert((int)src2 % 8 == 0);
    _nassert((int)src3 % 8 == 0);
    _nassert((int)dst1 % 8 == 0);
    _nassert((int)dst2 % 8 == 0);
    _nassert(width >= 32);
    _nassert(width % 4 == 0);

    cF2 = *src2;
    src2++;
    cF2 = ((uint64_t) _pack2(cF2, cF2)) << 32;

    // inner loop (4 pixels per iteration)
    #pragma MUST_ITERATE(7, , )
    for (width -= 4; width != 0; width -= 4)
    {
// A B C  --\ E0 E1
// D E F  --/ E2 E3
// G H I

        cB2 = _amem8_const(src1);
        src1 += 4;

        cnewD = _hill(cF2);
        cF2 = _mem8_const(src2);
        src2 += 4;
        cD2 = _itoll(_loll(cF2), cnewD);

        cH2 = _amem8_const(src3);
        src3 += 4;


        #define cB _loll(cB2)
        #define cD _loll(cD2)
        #define cF _loll(cF2)
        #define cH _loll(cH2)

        cE = _packlh2(cF, cD);

        C0 = _cmpeq2(cB, cH) | _cmpeq2(cD, cF);             // C0 = B == H || D == F
        tmp0 = _xpnd2(C0 | ~_cmpeq2(cD, cB));               // tmp0 = C0 || D != B
        tmp1 = _xpnd2(C0 | ~_cmpeq2(cB, cF));               // tmp1 = C0 || B != F

        E0 = (cE & tmp0) | (cB & ~tmp0);                    // E0 = tmp0 ? E : B
        E1 = (cE & tmp1) | (cB & ~tmp1);                    // E1 = tmp1 ? E : B

        _amem8(dst1) = _dpack2(E1, E0);
        dst1 += 4;


        tmp2 = _xpnd2(C0 | ~_cmpeq2(cD, cH));               // tmp2 = C0 || D != H
        tmp3 = _xpnd2(C0 | ~_cmpeq2(cH, cF));               // tmp3 = C0 || H != F

        E2 = (cE & tmp2) | (cH & ~tmp2);                    // E2 = tmp2 ? E : H
        E3 = (cE & tmp3) | (cH & ~tmp3);                    // E3 = tmp3 ? E : H

        _amem8(dst2) = _dpack2(E3, E2);
        dst2 += 4;

        #undef cH
        #undef cF
        #undef cD
        #undef cB


        #define cB _hill(cB2)
        #define cD _hill(cD2)
        #define cF _hill(cF2)
        #define cH _hill(cH2)

        cE = _packlh2(cF, cD);

        C0 = _cmpeq2(cB, cH) | _cmpeq2(cD, cF);             // C0 = B == H || D == F
        tmp0 = _xpnd2(C0 | ~_cmpeq2(cD, cB));               // tmp0 = C0 || D != B
        tmp1 = _xpnd2(C0 | ~_cmpeq2(cB, cF));               // tmp1 = C0 || B != F

        E0 = (cE & tmp0) | (cB & ~tmp0);                    // E0 = tmp0 ? E : B
        E1 = (cE & tmp1) | (cB & ~tmp1);                    // E1 = tmp1 ? E : B

        _amem8(dst1) = _dpack2(E1, E0);
        dst1 += 4;


        tmp2 = _xpnd2(C0 | ~_cmpeq2(cD, cH));               // tmp2 = C0 || D != H
        tmp3 = _xpnd2(C0 | ~_cmpeq2(cH, cF));               // tmp3 = C0 || H != F

        E2 = (cE & tmp2) | (cH & ~tmp2);                    // E2 = tmp2 ? E : H
        E3 = (cE & tmp3) | (cH & ~tmp3);                    // E3 = tmp3 ? E : H

        _amem8(dst2) = _dpack2(E3, E2);
        dst2 += 4;

        #undef cH
        #undef cF
        #undef cD
        #undef cB
    }

    // last 4 pixels
    {
        cB2 = _amem8_const(src1);
        //src1 += 4;

        cnewD = _hill(cF2);
        cF2 = _itoll(_pack2(src2[2], src2[2]), _mem4_const(src2));
        //src2 += 4;
        cD2 = _itoll(_loll(cF2), cnewD);

        cH2 = _amem8_const(src3);
        //src3 += 4;


        #define cB _loll(cB2)
        #define cD _loll(cD2)
        #define cF _loll(cF2)
        #define cH _loll(cH2)

        cE = _packlh2(cF, cD);

        C0 = _cmpeq2(cB, cH) | _cmpeq2(cD, cF);             // C0 = B == H || D == F
        tmp0 = _xpnd2(C0 | ~_cmpeq2(cD, cB));               // tmp0 = C0 || D != B
        tmp1 = _xpnd2(C0 | ~_cmpeq2(cB, cF));               // tmp1 = C0 || B != F

        E0 = (cE & tmp0) | (cB & ~tmp0);                    // E0 = tmp0 ? E : B
        E1 = (cE & tmp1) | (cB & ~tmp1);                    // E1 = tmp1 ? E : B

        _amem8(dst1) = _dpack2(E1, E0);
        dst1 += 4;


        tmp2 = _xpnd2(C0 | ~_cmpeq2(cD, cH));               // tmp2 = C0 || D != H
        tmp3 = _xpnd2(C0 | ~_cmpeq2(cH, cF));               // tmp3 = C0 || H != F

        E2 = (cE & tmp2) | (cH & ~tmp2);                    // E2 = tmp2 ? E : H
        E3 = (cE & tmp3) | (cH & ~tmp3);                    // E3 = tmp3 ? E : H

        _amem8(dst2) = _dpack2(E3, E2);
        dst2 += 4;

        #undef cH
        #undef cF
        #undef cD
        #undef cB


        #define cB _hill(cB2)
        #define cD _hill(cD2)
        #define cF _hill(cF2)
        #define cH _hill(cH2)

        cE = _packlh2(cF, cD);

        C0 = _cmpeq2(cB, cH) | _cmpeq2(cD, cF);             // C0 = B == H || D == F
        tmp0 = _xpnd2(C0 | ~_cmpeq2(cD, cB));               // tmp0 = C0 || D != B
        tmp1 = _xpnd2(C0 | ~_cmpeq2(cB, cF));               // tmp1 = C0 || B != F

        E0 = (cE & tmp0) | (cB & ~tmp0);                    // E0 = tmp0 ? E : B
        E1 = (cE & tmp1) | (cB & ~tmp1);                    // E1 = tmp1 ? E : B

        _amem8(dst1) = _dpack2(E1, E0);
        //dst1 += 4;


        tmp2 = _xpnd2(C0 | ~_cmpeq2(cD, cH));               // tmp2 = C0 || D != H
        tmp3 = _xpnd2(C0 | ~_cmpeq2(cH, cF));               // tmp3 = C0 || H != F

        E2 = (cE & tmp2) | (cH & ~tmp2);                    // E2 = tmp2 ? E : H
        E3 = (cE & tmp3) | (cH & ~tmp3);                    // E3 = tmp3 ? E : H

        _amem8(dst2) = _dpack2(E3, E2);
        //dst2 += 4;

        #undef cH
        #undef cF
        #undef cD
        #undef cB
    }
}

void _scale2x_16_16(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    _scale2x_16_16_line(src, src, src + (srcstride / 2), dst, dst + (dststride / 2), width);

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    for (height -= 2; height != 0; height--)
    {
        _scale2x_16_16_line(src - (srcstride / 2), src, src + (srcstride / 2), dst, dst + (dststride / 2), width);

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    _scale2x_16_16_line(src - (srcstride / 2), src, src, dst, dst + (dststride / 2), width);
}

void _scale2x_16_16_srcdma(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict tmpsrc1;
    uint16_t * restrict tmpsrc2;
    uint16_t * restrict tmpsrc3;
    uint16_t * restrict tmpsrc4;

    tmpsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    syscalls.qdma_copy2d(QDMA_0, tmpsrc1, 2*width, src, srcstride, 2*width, 2);

    tmpsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    tmpsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);
    tmpsrc4 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src + srcstride, 2*width);

    _scale2x_16_16_line(tmpsrc1, tmpsrc1, tmpsrc2, dst, dst + (dststride / 2), width);

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    for (height -= 3; height != 0; height--)
    {
        syscalls.qdma_wait(QDMA_0);
        syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src + srcstride, 2*width);
        _scale2x_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, dst, dst + (dststride / 2), width);

        {
            uint16_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    syscalls.qdma_wait(QDMA_0);
    _scale2x_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, dst, dst + (dststride / 2), width);

    //src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    _scale2x_16_16_line(tmpsrc2, tmpsrc3, tmpsrc3, dst, dst + (dststride / 2), width);
}

void _scale2x_16_16_dstdma(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    tmpdst1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    tmpdst2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);

    _scale2x_16_16_line(src, src, src + (srcstride / 2), tmpdst2, tmpdst2 + (2*width), width);

    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    for (height -= 2; height != 0; height--)
    {
        _scale2x_16_16_line(src - (srcstride / 2), src, src + (srcstride / 2), tmpdst1, tmpdst1 + (2*width), width);

        syscalls.qdma_wait(QDMA_1);
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);

        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    _scale2x_16_16_line(src - (srcstride / 2), src, src, tmpdst1, tmpdst1 + (2*width), width);

    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_wait(QDMA_0);
}

void _scale2x_16_16_dma(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict tmpsrc1;
    uint16_t * restrict tmpsrc2;
    uint16_t * restrict tmpsrc3;
    uint16_t * restrict tmpsrc4;
    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    tmpsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    syscalls.qdma_copy2d(QDMA_0, tmpsrc1, 2*width, src, srcstride, 2*width, 2);

    tmpsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    tmpsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);
    tmpsrc4 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);
    tmpdst1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);
    tmpdst2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 16*width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src + srcstride, 2*width);

    _scale2x_16_16_line(tmpsrc1, tmpsrc1, tmpsrc2, tmpdst2, tmpdst2 + (2*width), width);

    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    for (height -= 3; height != 0; height--)
    {
        syscalls.qdma_wait(QDMA_0);
        syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src + srcstride, 2*width);
        _scale2x_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst1, tmpdst1 + (2*width), width);

        syscalls.qdma_wait(QDMA_1);
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);

        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    syscalls.qdma_wait(QDMA_0);
    _scale2x_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst1, tmpdst1 + (2*width), width);

    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);

    //src += (srcstride / 2);
    dst += 2 * (dststride / 2);

    _scale2x_16_16_line(tmpsrc2, tmpsrc3, tmpsrc3, tmpdst2, tmpdst2 + (2*width), width);

    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst2, 4*width, 4*width, 2);
    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_wait(QDMA_0);
}

void _scale2x_16_16_dmalink(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 4 == 0);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict tmpsrc1;
    uint16_t * restrict tmpsrc2;
    uint16_t * restrict tmpsrc3;
    uint16_t * restrict tmpsrc4;
    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    tmpsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    syscalls.qdma_copy2d(QDMA_0, tmpsrc1, 2*width, src, srcstride, 2*width, 2);

    tmpsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    tmpsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);
    tmpsrc4 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);
    tmpdst1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);
    tmpdst2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 16*width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src + srcstride, 2*width);

    _scale2x_16_16_line(tmpsrc1, tmpsrc1, tmpsrc2, tmpdst2, tmpdst2 + (2*width), width);

    src += (srcstride / 2);

    for (height -= 3; height != 0; height--)
    {
        syscalls.qdma_wait(QDMA_0);
        syscalls.qdma_link2d(QDMA_0, 8, 0xFFFF, dst, dststride, tmpdst2, 4*width, 4*width, 2);
        syscalls.qdma_link1d(QDMA_0, QDMA_0, 8, tmpsrc4, src + srcstride, 2*width);

        _scale2x_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst1, tmpdst1 + (2*width), width);

        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }

        src += (srcstride / 2);
        dst += 2 * (dststride / 2);
    }

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    _scale2x_16_16_line(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst1, tmpdst1 + (2*width), width);

    dst += 2 * (dststride / 2);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);

    _scale2x_16_16_line(tmpsrc2, tmpsrc3, tmpsrc3, tmpdst2, tmpdst2 + (2*width), width);

    dst += 2 * (dststride / 2);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    syscalls.qdma_wait(QDMA_0);
}


void _scale2x_8_16(const uint8_t * restrict src, uint16_t * restrict dst, const uint16_t * restrict pal, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert((int)pal % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 8 == 0);
    _nassert(srcstride >= width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict newsrc1;
    uint16_t * restrict newsrc2;
    uint16_t * restrict newsrc3;

    newsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    newsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    newsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);

    _normal1x_8_16_line(src, newsrc1, pal, width);
    _normal1x_8_16_line(src + srcstride, newsrc2, pal, width);

    _scale2x_16_16_line(newsrc1, newsrc1, newsrc2, dst, dst + (dststride / 2), width);

    for (height -= 2; height != 0; height--)
    {
        src += srcstride;
        dst += 2 * (dststride / 2);

        _normal1x_8_16_line(src + srcstride, newsrc3, pal, width);
        _scale2x_16_16_line(newsrc1, newsrc2, newsrc3, dst, dst + (dststride / 2), width);

        {
            uint16_t * restrict newsrc;
            newsrc = newsrc1;
            newsrc1 = newsrc2;
            newsrc2 = newsrc3;
            newsrc3 = newsrc;
        }
    }

    dst += 2 * (dststride / 2);

    _scale2x_16_16_line(newsrc1, newsrc2, newsrc2, dst, dst + (dststride / 2), width);
}

void _scale2x_8_16_srcdma(const uint8_t * restrict src, uint16_t * restrict dst, const uint16_t * restrict pal, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert((int)pal % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 8 == 0);
    _nassert(srcstride >= width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint8_t * restrict tmpsrc1;
    uint8_t * restrict tmpsrc2;
    uint8_t * restrict tmpsrc3;
    uint8_t * restrict tmpsrc4;
    uint16_t * restrict newsrc1;
    uint16_t * restrict newsrc2;
    uint16_t * restrict newsrc3;


    tmpsrc1 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc1, src, width);

    tmpsrc2 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 1*width);
    tmpsrc3 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    tmpsrc4 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 3*width);

    newsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);
    newsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);
    newsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc2, src + srcstride, width);
    _normal1x_8_16_line(tmpsrc1, newsrc1, pal, width);

    syscalls.qdma_wait(QDMA_0);
    _normal1x_8_16_line(tmpsrc2, newsrc2, pal, width);

    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src + 2*srcstride, width);
    _scale2x_16_16_line(newsrc1, newsrc1, newsrc2, dst, dst + (dststride / 2), width);

    for (height -= 3; height != 0; height--)
    {
        src += srcstride;
        dst += 2 * (dststride / 2);

        syscalls.qdma_wait(QDMA_0);
        _normal1x_8_16_line(tmpsrc3, newsrc3, pal, width);
        syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src + 2*srcstride, width);
        _scale2x_16_16_line(newsrc1, newsrc2, newsrc3, dst, dst + (dststride / 2), width);

        {
            uint16_t * restrict newsrc;
            newsrc = newsrc1;
            newsrc1 = newsrc2;
            newsrc2 = newsrc3;
            newsrc3 = newsrc;
        }
        {
            uint8_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }
    }


    dst += 2 * (dststride / 2);

    syscalls.qdma_wait(QDMA_0);
    _normal1x_8_16_line(tmpsrc3, newsrc3, pal, width);
    _scale2x_16_16_line(newsrc1, newsrc2, newsrc3, dst, dst + (dststride / 2), width);

    dst += 2 * (dststride / 2);

    _scale2x_16_16_line(newsrc2, newsrc3, newsrc3, dst, dst + (dststride / 2), width);
}

void _scale2x_8_16_dstdma(const uint8_t * restrict src, uint16_t * restrict dst, const uint16_t * restrict pal, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert((int)pal % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 8 == 0);
    _nassert(srcstride >= width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint16_t * restrict newsrc1;
    uint16_t * restrict newsrc2;
    uint16_t * restrict newsrc3;
    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;


    newsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    newsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    newsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);

    tmpdst1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);
    tmpdst2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 14*width);

    _normal1x_8_16_line(src, newsrc1, pal, width);
    _normal1x_8_16_line(src + srcstride, newsrc2, pal, width);

    _scale2x_16_16_line(newsrc1, newsrc1, newsrc2, tmpdst2, tmpdst2 + (2*width), width);

    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    for (height -= 2; height != 0; height--)
    {
        src += srcstride;
        dst += 2 * (dststride / 2);

        _normal1x_8_16_line(src + srcstride, newsrc3, pal, width);
        _scale2x_16_16_line(newsrc1, newsrc2, newsrc3, tmpdst1, tmpdst1 + (2*width), width);

        syscalls.qdma_wait(QDMA_1);
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);

        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict newsrc;
            newsrc = newsrc1;
            newsrc1 = newsrc2;
            newsrc2 = newsrc3;
            newsrc3 = newsrc;
        }
    }

    dst += 2 * (dststride / 2);

    _scale2x_16_16_line(newsrc1, newsrc2, newsrc2, tmpdst1, tmpdst1 + (2*width), width);

    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_wait(QDMA_0);
}

void _scale2x_8_16_dma(const uint8_t * restrict src, uint16_t * restrict dst, const uint16_t * restrict pal, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert((int)pal % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 8 == 0);
    _nassert(srcstride >= width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint8_t * restrict tmpsrc1;
    uint8_t * restrict tmpsrc2;
    uint8_t * restrict tmpsrc3;
    uint8_t * restrict tmpsrc4;
    uint16_t * restrict newsrc1;
    uint16_t * restrict newsrc2;
    uint16_t * restrict newsrc3;
    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;


    tmpsrc1 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc1, src, width);

    tmpsrc2 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 1*width);
    tmpsrc3 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    tmpsrc4 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 3*width);

    newsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);
    newsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);
    newsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);

    tmpdst1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 10*width);
    tmpdst2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 18*width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc2, src + srcstride, width);
    _normal1x_8_16_line(tmpsrc1, newsrc1, pal, width);

    syscalls.qdma_wait(QDMA_0);
    _normal1x_8_16_line(tmpsrc2, newsrc2, pal, width);

    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src + 2*srcstride, width);
    _scale2x_16_16_line(newsrc1, newsrc1, newsrc2, tmpdst2, tmpdst2 + (2*width), width);

    for (height -= 3; height != 0; height--)
    {
        syscalls.qdma_wait(QDMA_0);
        _normal1x_8_16_line(tmpsrc3, newsrc3, pal, width);

        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);

        src += srcstride;
        dst += 2 * (dststride / 2);

        syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src + 2*srcstride, width);
        _scale2x_16_16_line(newsrc1, newsrc2, newsrc3, tmpdst1, tmpdst1 + (2*width), width);

        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict newsrc;
            newsrc = newsrc1;
            newsrc1 = newsrc2;
            newsrc2 = newsrc3;
            newsrc3 = newsrc;
        }
        {
            uint8_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }

        syscalls.qdma_wait(QDMA_1);
    }

    syscalls.qdma_wait(QDMA_0);
    _normal1x_8_16_line(tmpsrc3, newsrc3, pal, width);

    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    dst += 2 * (dststride / 2);

    _scale2x_16_16_line(newsrc1, newsrc2, newsrc3, tmpdst1, tmpdst1 + (2*width), width);

    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);

    dst += 2 * (dststride / 2);

    _scale2x_16_16_line(newsrc2, newsrc3, newsrc3, tmpdst2, tmpdst2 + (2*width), width);

    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst2, 4*width, 4*width, 2);
    syscalls.qdma_wait(QDMA_1);
    syscalls.qdma_wait(QDMA_0);
}

void _scale2x_8_16_dmalink(const uint8_t * restrict src, uint16_t * restrict dst, const uint16_t * restrict pal, int width, int srcstride, int dststride, int height)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert((int)pal % 8 == 0);
    _nassert(width >= 32);
    _nassert(height >= 4);
    _nassert(width % 8 == 0);
    _nassert(srcstride >= width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);

    uint8_t * restrict tmpsrc1;
    uint8_t * restrict tmpsrc2;
    uint8_t * restrict tmpsrc3;
    uint8_t * restrict tmpsrc4;
    uint16_t * restrict newsrc1;
    uint16_t * restrict newsrc2;
    uint16_t * restrict newsrc3;
    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;


    tmpsrc1 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc1, src, width);

    tmpsrc2 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 1*width);
    tmpsrc3 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 2*width);
    tmpsrc4 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 3*width);

    newsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 4*width);
    newsrc2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 6*width);
    newsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8*width);

    tmpdst1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 10*width);
    tmpdst2 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 18*width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy1d(QDMA_0, tmpsrc2, src + srcstride, width);
    _normal1x_8_16_line(tmpsrc1, newsrc1, pal, width);

    syscalls.qdma_wait(QDMA_0);
    _normal1x_8_16_line(tmpsrc2, newsrc2, pal, width);

    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src + 2*srcstride, width);
    _scale2x_16_16_line(newsrc1, newsrc1, newsrc2, tmpdst2, tmpdst2 + (2*width), width);

    for (height -= 3; height != 0; height--)
    {
        syscalls.qdma_wait(QDMA_0);
        _normal1x_8_16_line(tmpsrc3, newsrc3, pal, width);

        syscalls.qdma_link2d(QDMA_0, 8, 0xFFFF, dst, dststride, tmpdst2, 4*width, 4*width, 2);

        src += srcstride;
        dst += 2 * (dststride / 2);

        syscalls.qdma_link1d(QDMA_0, QDMA_0, 8, tmpsrc4, src + 2*srcstride, width);
        _scale2x_16_16_line(newsrc1, newsrc2, newsrc3, tmpdst1, tmpdst1 + (2*width), width);

        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict newsrc;
            newsrc = newsrc1;
            newsrc1 = newsrc2;
            newsrc2 = newsrc3;
            newsrc3 = newsrc;
        }
        {
            uint8_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }
    }

    syscalls.qdma_wait(QDMA_0);
    _normal1x_8_16_line(tmpsrc3, newsrc3, pal, width);

    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    dst += 2 * (dststride / 2);

    _scale2x_16_16_line(newsrc1, newsrc2, newsrc3, tmpdst1, tmpdst1 + (2*width), width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);

    dst += 2 * (dststride / 2);

    _scale2x_16_16_line(newsrc2, newsrc3, newsrc3, tmpdst2, tmpdst2 + (2*width), width);

    syscalls.qdma_wait(QDMA_0);
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst2, 4*width, 4*width, 2);

    syscalls.qdma_wait(QDMA_0);
}

